#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tb.h"



void tb_initialize(struct tb_table* t)
{
   t->pCol = 0;
   t->pRow = 0;
   t->iColNum = 0;
   t->iRowNum = 0;
}



static void halt(unsigned int Code)
{
   printf("ERROR: ");
   switch (Code)
   {
      case 1 : printf("Memory allocation error"); break;
      case 2 : printf("Table hasn't been described"); break;
      case 3 : printf("Illegal index"); break;
      case 4 : printf("Wrong type"); break;
   }
   printf("\n");
   exit(Code);
}



static void add_row(struct tb_table* t, unsigned int n)
{
   unsigned int i;
   t->pRow[n] = (union tb_cell*)malloc(t->iColNum * sizeof(union tb_cell));
   if (t->pRow[n] == 0) halt(1);
   for (i = 0; i < t->iColNum; i++)
   if (t->pCol[i].type == tb_type_char)
   {
      t->pRow[n][i].pChar = (char*)malloc(t->pCol[i].size);
      if (t->pRow[n][i].pChar == 0) halt(1);
   }
}



static void del_row(struct tb_table* t, unsigned int n)
{
   unsigned int i;
   for (i = 0; i < t->iColNum; i++)
   if (t->pCol[i].type == tb_type_char)
   free(t->pRow[n][i].pChar);
   free(t->pRow[n]);
}



void tb_clear(struct tb_table* t)
{
   unsigned int i;
   if (t->iColNum == 0) return;
   for (i = 0; i < t->iRowNum; i++) del_row(t, i);
   free(t->pRow);
   t->pRow = 0;
   t->iRowNum = 0;
}



void tb_discard(struct tb_table* t)
{
   if (t->iColNum == 0) return;
   tb_clear(t);
   free(t->pCol);
   t->pCol = 0;
   t->iColNum = 0;
}



void tb_describe(struct tb_table* t, struct tb_column* col, unsigned int n, unsigned int occurs)
{
   unsigned int i;
   if (n == 0) halt(3);
   if (t->iColNum != 0) tb_discard(t);
   t->pCol = (struct tb_column*)malloc(n * sizeof(struct tb_column));
   if (t->pCol == 0) halt(1);
   for (i = 0; i < n; i++) t->pCol[i] = col[i];
   t->iColNum = n;
}



unsigned int tb_insert(struct tb_table* t, unsigned int n)
{
   if (n > t->iRowNum) halt(3);
   if (t->iColNum == 0) halt(2);
   t->pRow = (union tb_cell**)realloc(t->pRow, (t->iRowNum+1) * sizeof(union tb_cell*));
   if (t->pRow == 0) halt(1);
   memmove(t->pRow + n + 1, t->pRow + n, (t->iRowNum-n) * sizeof(union tb_cell*));
   add_row(t, n);
   t->iRowNum++;
   return n;
}



unsigned int tb_append(struct tb_table* t)
{
   if (t->iColNum == 0) halt(2);
   t->pRow = (union tb_cell**)realloc(t->pRow, (t->iRowNum+1) * sizeof(union tb_cell*));
   if (t->pRow == 0) halt(1);
   add_row(t, t->iRowNum);
   return t->iRowNum++;
}



unsigned int tb_getnrows(struct tb_table* t)
{
   return t->iRowNum;
}



int tb_remove(struct tb_table* t, unsigned int n)
{
   if (n >= t->iRowNum) halt(3);
   del_row(t, n);
   t->iRowNum--;
   memmove(t->pRow + n, t->pRow + n + 1, (t->iRowNum-n) * sizeof(union tb_cell*));
   t->pRow = (union tb_cell**)realloc(t->pRow, (t->iRowNum) * sizeof(union tb_cell*));
   return (t->iRowNum != 0) ? n : -1;
}



int tb_setint(struct tb_table* t, unsigned int n, unsigned int m, int v)
{
   if ((n >= t->iRowNum) || (m >= t->iColNum)) halt(3);
   if (t->pCol[m].type != tb_type_int) halt(4);
   t->pRow[n][m].iInt = v;
   return v;
}



int tb_getint(struct tb_table* t, unsigned int n, unsigned int m)
{
   if ((n >= t->iRowNum) || (m >= t->iColNum)) halt(3);
   if (t->pCol[m].type != tb_type_int) halt(4);
   return t->pRow[n][m].iInt;
}



double tb_setdouble(struct tb_table* t, unsigned int n, unsigned int m, double v)
{
   if ((n >= t->iRowNum) || (m >= t->iColNum)) halt(3);
   if (t->pCol[m].type != tb_type_double) halt(4);
   t->pRow[n][m].dDouble = v;
   return v;
}



double tb_getdouble(struct tb_table* t, unsigned int n, unsigned int m)
{
   if ((n >= t->iRowNum) || (m >= t->iColNum)) halt(3);
   if (t->pCol[m].type != tb_type_double) halt(4);
   return t->pRow[n][m].dDouble;
}



char* tb_setchar(struct tb_table* t, unsigned int n, unsigned int m, char* v)
{
   if ((n >= t->iRowNum) || (m >= t->iColNum)) halt(3);
   if (t->pCol[m].type != tb_type_char) halt(4);
   strncpy(t->pRow[n][m].pChar, v, t->pCol[m].size);
   t->pRow[n][m].pChar[t->pCol[m].size-1] = '\0';
   return v;
}



char* tb_getchar(struct tb_table* t, unsigned int n, unsigned int m, char* v, unsigned int size)
{
   if ((n >= t->iRowNum) || (m >= t->iColNum)) halt(3);
   if (t->pCol[m].type != tb_type_char) halt(4);
   strncpy(v, t->pRow[n][m].pChar, size);
   v[size-1] = '\0';
   return v;
}



int tb_findcol(struct tb_table* t, const char* name)
{
   unsigned int i;
   for (i = 0; i < t->iColNum; i++)
   if (!strcmp(t->pCol[i].name, name)) return i;
   return -1;
}



unsigned int tb_match(struct tb_table* t, struct tb_column* col, unsigned int n, unsigned int* pos, int exact)
{
   unsigned int i,j;
   unsigned int F1,F2;

   if (t->iColNum == 0) halt(2);

   for (F2 = 1, i = 0; (i < n) && (F2 != 0); i++)
   {
      for (F1 = 1, j = 0; j < t->iColNum; j++)
      {
	 F1 &= !strcmp(t->pCol[j].name, col[i].name);
	 F1 &= (t->pCol[j].type == col[i].type);
	 if (col[i].type == tb_type_char) F1 &= (t->pCol[j].size == col[i].size);
	 if (F1 != 0) break;
      }
      if ((F1 != 0) && (pos != 0)) pos[i] = j;
      F2 &= F1;
   }
   if (exact != 1) F2 &= (t->iColNum == n);
   return F2;
}



void tb_getcolumn(struct tb_table* t, unsigned int n, struct tb_column* col)
{
   if (n >= t->iColNum) halt(3);
   *col = t->pCol[n];
}



static int row_cmp(void* c1, void* c2, enum tb_type type)
{
   if (type == tb_type_int)
   {
       if (*(int*)c1 < *(int*)c2) return -1;
       if (*(int*)c1 > *(int*)c2) return 1;
       return 0;
   }
   if (type == tb_type_double)
   {
       if (*(double*)c1 < *(double*)c2) return -1;
       if (*(double*)c1 > *(double*)c2) return 1;
       return 0;
   }
   return strcmp(*(char**)c1, *(char**)c2);
}



void tb_sort(struct tb_table* t, unsigned int* p, unsigned int n, int* o)
{
   int F;
   union tb_cell* pCell;
   unsigned int i,j,k,m;

   for (m = 0; m < n; m++)
   if (p[m] >= t->iColNum) halt(3);

//    ()

   for (k = t->iRowNum/2; k > 0; k /= 2)
   for (i = k; i < t->iRowNum; i++)
   for (j = i; j >= k; j -= k)
   {
      for (F = 0, m = 0; (m < n) && (F == 0); m++)
      F |= row_cmp(&t->pRow[j-k][p[m]], &t->pRow[j][p[m]], t->pCol[p[m]].type);
      if (F*(o[m-1] | 1) < 0) break;
      pCell = t->pRow[j];
      t->pRow[j] = t->pRow[j-k];
      t->pRow[j-k] = pCell;
   }

//   

/*
   for (i = 0; i < t->iRowNum-1; i++)
   {
      for (k = i, j = i+1; j < t->iRowNum; j++)
      {
	 for (F = 0, m = 0; (m < n) && (F == 0); m++)
	 F |= row_cmp(&t->pRow[j][p[m]], &t->pRow[k][p[m]], t->pCol[p[m]].type);
	 if (F*(o[m-1] | 1) < 0) k = j;
      }
      pCell = t->pRow[i];
      t->pRow[i] = t->pRow[k];
      t->pRow[k] = pCell;
   }
*/

}



int tb_bsearch(struct tb_table* t, unsigned int* p, unsigned int n, void* d[])
{
   int F;
   unsigned int i,m;
   unsigned int L,R;

   for (m = 0; m < n; m++)
   if (p[m] >= t->iColNum) halt(3);

   L = 0;
   R = t->iRowNum - 1;
   while (L <= R)
   {
      i = (L+R)/2;
      for (F = 0, m = 0; (m < n) && (F == 0); m++)
      F |= row_cmp(&t->pRow[i][p[m]], d[m], t->pCol[p[m]].type);
      if (F == 0) break;
      if (F < 0) L=i+1; else R=i-1;
//    if (F*(o[m-1] | 1) < 0) L=i+1; else R=i-1;
   }
   return (F == 0) ? i : -1;
}
